#include "GraphicsPipeline.hpp"

[[nodiscard]]
static auto
    load_shader_module(const vk::Device t_device, const std::filesystem::path& t_filepath)
        -> vk::UniqueShaderModule
{
    std::ifstream file{ t_filepath, std::ios::binary | std::ios::in | std::ios::ate };

    const std::streamsize file_size = file.tellg();
    if (file_size == -1) {
        throw std::runtime_error{
            std::format("Failed to open file: {}", t_filepath.generic_string())
        };
    }

    std::vector<char> buffer(static_cast<size_t>(file_size));

    file.seekg(0, std::ios::beg);
    file.read(buffer.data(), file_size);
    file.close();

    const vk::ShaderModuleCreateInfo create_info{
        .codeSize = static_cast<size_t>(file_size),
        .pCode    = reinterpret_cast<uint32_t*>(buffer.data())
    };

    return t_device.createShaderModuleUnique(create_info);
}

auto create_pipeline_layout(const vk::Device device) -> vk::UniquePipelineLayout
{
    constexpr static vk::PipelineLayoutCreateInfo create_info{};

    return device.createPipelineLayoutUnique(create_info);
}

auto build_graphics_pipeline(
    const vk::Device         device,
    const vk::PipelineLayout layout,
    const vk::RenderPass     render_pass
) -> vk::UniquePipeline
{
    const vk::UniqueShaderModule vertex_shader{
        load_shader_module(device, "shaders/triangle.frag")
    };
    const vk::UniqueShaderModule fragment_shader{
        load_shader_module(device, "shaders/triangle.frag")
    };
    const std::array shader_stages{
        vk::PipelineShaderStageCreateInfo{   .stage  = vk::ShaderStageFlagBits::eVertex,
                                          .module = vertex_shader.get(),
                                          .pName  = "main" },
        vk::PipelineShaderStageCreateInfo{ .stage  = vk::ShaderStageFlagBits::eFragment,
                                          .module = fragment_shader.get(),
                                          .pName  = "main" }
    };

    constexpr static vk::PipelineVertexInputStateCreateInfo
        vertex_input_state_create_info{};

    constexpr static vk::PipelineInputAssemblyStateCreateInfo
        input_assembly_state_create_info{
            .topology = vk::PrimitiveTopology::eTriangleList,
        };

    constexpr static vk::PipelineViewportStateCreateInfo viewport_state_create_info{
        .viewportCount = 1,
        .scissorCount  = 1,
    };

    constexpr static vk::PipelineRasterizationStateCreateInfo
        rasterization_state_create_info{
            .polygonMode = vk::PolygonMode::eFill,
            .cullMode    = vk::CullModeFlagBits::eBack,
            .frontFace   = vk::FrontFace::eCounterClockwise,
            .lineWidth   = 1.f,
        };

    constexpr static vk::PipelineMultisampleStateCreateInfo multisample_state_create_info{
        .rasterizationSamples = vk::SampleCountFlagBits::e1,
    };

    constexpr static vk::PipelineDepthStencilStateCreateInfo
        depth_stencil_state_create_info{
            .depthTestEnable       = true,
            .depthWriteEnable      = true,
            .depthCompareOp        = vk::CompareOp::eLess,
            .depthBoundsTestEnable = false,
            .stencilTestEnable     = false,
        };

    constexpr static vk::PipelineColorBlendAttachmentState color_blend_attachment_state{
        .blendEnable    = vk::False,
        .colorWriteMask = vk::ColorComponentFlagBits::eR | vk::ColorComponentFlagBits::eG
                        | vk::ColorComponentFlagBits::eB | vk::ColorComponentFlagBits::eA,
    };
    constexpr static vk::PipelineColorBlendStateCreateInfo color_blend_state_create_info{
        .attachmentCount = 1,
        .pAttachments    = &color_blend_attachment_state,
    };

    constexpr static std::array dynamic_states{
        vk::DynamicState::eViewport,
        vk::DynamicState::eScissor,
    };
    constexpr static vk::PipelineDynamicStateCreateInfo dynamic_state_create_info{
        .dynamicStateCount = static_cast<uint32_t>(dynamic_states.size()),
        .pDynamicStates    = dynamic_states.data(),
    };

    const vk::GraphicsPipelineCreateInfo create_info{
        .stageCount          = static_cast<uint32_t>(shader_stages.size()),
        .pStages             = shader_stages.data(),
        .pVertexInputState   = &vertex_input_state_create_info,
        .pInputAssemblyState = &input_assembly_state_create_info,
        .pViewportState      = &viewport_state_create_info,
        .pRasterizationState = &rasterization_state_create_info,
        .pMultisampleState   = &multisample_state_create_info,
        .pDepthStencilState  = &depth_stencil_state_create_info,
        .pColorBlendState    = &color_blend_state_create_info,
        .pDynamicState       = &dynamic_state_create_info,
        .layout              = layout,
        .renderPass          = render_pass,
    };

    return device.createGraphicsPipelineUnique(nullptr, create_info).value;
}
